/* depends on:
	jquery core 1.4.2 or later <http://jquery.com>
	jquery scrollfollow plugin <http://kitchen.net-perspective.com/>
	jquery placeheld plugin <http://plugins.jquery.com/project/placeheld>
*/


// Array Remove - By John Resig (MIT Licensed)
Array.prototype.remove = function(from, to) {
  var rest = this.slice((to || from) + 1 || this.length);
  this.length = from < 0 ? this.length + from : from;
  return this.push.apply(this, rest);
};


var harvest_results;
var harvest_filters_list;


function Filter (dom_node) {
	/* Attaches to a .filter object in the document. Handles events
	   specific to each type of filter and provides some extra data
	   and utilities. */
	
	var filter = this; /* so we can access this from event handlers */
	
	$(dom_node).data('filter',this); /* so we can get here from the JQuery object */
	
	this.full_name = $(dom_node).attr('data-filter-fullname');
	this.parent_filter_node = $(dom_node).parent().closest('.filter');
	this.container = $(dom_node).closest('li'); /* the list item in the parent that this belongs to */
	this.value_node = $(dom_node).children('.filter-value');
	
	
	this.is_class = function (class_name) {
		return $(dom_node).hasClass(class_name);
	}
	
	
	this.get_value_serialized = function () {
		/* Returns the value of this filter, serialized in the same
		   format we would use in Harvest's Django end.
		   Does nothing by default. Implement this in each
		   "class"-specific initializer below. */
		
		return null;
	}
	
	this.value_changed = function () {
		/* Should be called when the filter's value has changed.
		   This posts the new value to the global Results object */
		
		value = this.get_value_serialized();
		if ( value != null ) {
			harvest_results.post_query(this.full_name, value);
		}
	}
	
	
	/********/
	/* Ugly hacks for talking to the parent node */
	
	this.get_parent_filtergroup = function () {
		/* Returns the FilterGroup that is a parent of this filter.
		   Returns null if the parent doesn't exist or isn't a FilterGroup. */
		parent = null;
		if ( this.parent_filter_node ) {
			parent_test = this.parent_filter_node.data('filter');
			if ( parent_test && parent_test.is_class('filtergroup') ) parent = parent_test;
		} 
		return parent;
	}
	
	
	this.force_enable_prelight_add = function () {
		/* Tells the parent (if there is one) to create a prelight for
		   enabling this item.
		   We assume the parent is a FilterGroup, which is probably right. */
		parent_filter = this.get_parent_filtergroup();
		if ( parent_filter && ! parent_filter.item_is_selected(this.container) ) {
			/* only do this if we weren't already selected */
			parent_filter.toggle_item_prelight_add(this.container);
		}
	}
	
	this.force_enable_prelight_clear = function () {
		/* Tells the parent (if there is one) to remove the prelight for
		   enabling this item.
		   We assume the parent is a FilterGroup. */
		parent_filter = this.get_parent_filtergroup()
		
		if ( parent_filter ) parent_filter.toggle_item_prelight_clear(this.container);
	}
	
	
	this.force_enable = function () {
		/* Tells the parent (if there is one) to enable this item,
		   because it said so!
		   We assume the parent is a FilterGroup. */
		parent_filter = this.get_parent_filtergroup();
		if ( parent_filter && ! parent_filter.item_is_selected(this.container) ) {
			parent_filter.enable_item(this.container);
		}
	}
	
	
	/* "Signals" */
	this.enabled = function() {void(0)} /* called when this filter is enabled */
	this.disabled = function() {void(0)} /* called when this filter is disabled */
	
	
	/********/
	/* "Class"-specific initialization stuff. Setting up functions and
	   handlers in filter.value_node */
	
	if ( this.is_class('choicefilter') ) {
		var list_items = this.value_node.children('ul').children('li');
		
		
		/* "Signals" */
		this.item_enabled = function(item) {void(0)} /* called after enabling an item */
		this.item_disabled = function(item) {void(0)} /* called after disabling an item */
		
		
		this.get_value_serialized = function () {
			var item_ids = Array();
			/* we add the data-item-id attribute of all selected items */
			list_items.each(function () {
				if ($(this).attr('data-selected') != null) {
					item_ids.push($(this).attr('data-item-id'));
				}
			} );
			return item_ids.join(',');
		}
		
		
		this.item_is_selected = function (item) {
			return ( item.attr('data-selected') != null );
		}
		
		this.toggle_item_prelight_add = function (item) {
			if ( this.item_is_selected(item) ) {
				item.addClass('prelight-off');
			} else {
				item.addClass('prelight-on');
				item.children('.checkbox').html('&#10003;'); /* decoration :) */
				this.force_enable_prelight_add();
			}
		}
		
		this.toggle_item_prelight_clear = function (item) {
			/* remove the decoration in the checkbox if we added it */
			if ( ! this.item_is_selected(item) ) {
				item.children('.checkbox').html('');
			}
			
			item.removeClass('prelight-on');
			item.removeClass('prelight-off');
			this.force_enable_prelight_clear();
		}
		
		
		this.enable_item = function (item) {
			item.attr('data-selected','');
			item.children('.checkbox').html('&#10003;');
			
			this.force_enable();
			this.value_changed();
			
			this.item_enabled(item);
		}
		
		this.disable_item = function (item) {
			item.removeAttr('data-selected');
			item.children('.checkbox').html('');
			
			this.value_changed();
			
			this.item_disabled(item);
		}
		
		this.toggle_item = function (item) {
			if ( this.item_is_selected(item) ) {
				this.disable_item(item);
			} else {
				this.enable_item(item);
			}
		}
		
		
		this.get_item_toggle_node = function (item) {
			/* returns the node(s) used to toggle the given item in this filter */
			return item.children('a.item-toggle');
		}
		
		
		/* "Subclass" setup goes here because we need to do overrides before any movement happens */
		if ( this.is_class('filtergroup') ) {
			this.item_enabled = function (item) {
				/* Tell child it has been enabled. It might do something cool! */
				child_filter = item.children('.filter').data('filter');
				child_filter.enabled();
			}
			
			this.item_disabled = function (item) {
				/* Tell child it has been disabled. */
				child_filter = item.children('.filter').data('filter');
				child_filter.disabled();
			}
		
			this.get_item_toggle_node = function (item) {
				return item.children('.filter').children('.filter-label').children('a.item-toggle');
			}
		}
		
		
		/* setup the event handlers we described on all the items */
		list_items.each(function () {
			var item = $(this);
			toggle_node = filter.get_item_toggle_node(item);
			toggle_node.bind('mouseenter focusin', function () { filter.toggle_item_prelight_add(item); } );
			toggle_node.bind('mouseleave focusout', function () { filter.toggle_item_prelight_clear(item); } );
			
			toggle_node.bind('click', function () {
				filter.toggle_item(item);
				filter.toggle_item_prelight_clear(item);
				
				return false; /* stops link from being followed */
			} );
		});
	}
	
	
	if ( this.is_class('editfilter') ) {
		var edit_field = this.value_node.children('input');
		
		
		this.edit_field_changed = function() {
			this.force_enable();
			this.value_changed();
		}
		
		this.get_value_serialized = function () {
			return escape(edit_field.val());
		}
		
		this.enabled = function () {
			edit_field.focus();
		}
		
		/* TODO: add a clear button to this element, like Webkit does for input type="search" */ 
		/* we could handle 'change' instead of 'input' to only check value when typing has stopped */
		edit_field.bind('input', function () { filter.edit_field_changed(); } );
	}
	
	
	/* send initial state to Results object so it can behave neatly */
	value = this.get_value_serialized();
	if ( value != null ) {
		harvest_results.update_current_query(this.full_name, value);
	}
}


function Results (dom_node) {
	/* Attached to the single #results pane in the document. Receives
	   new query parameters from Filter objects, grabs and displays new
	   results. */
	
	/* The current_query variable is set with the current querystring,
	   but some filters could already have been set with Harvest's
	   default parameters while the querystring doesn't mention them.
	   On its own, this isn't a problem, but it may lead to some odd
	   behaviour. We should fix it by setting current_query to include
	   all of the defaults in Harvest's Django portion. */
	
	var results = this;
	
	this.query_countdown = 0;
	
	this.current_query = {}; /* we need to figure out what this is without using the querystring */
	this.future_query = {};
	
	this.results_pane = $(dom_node);
	this.results_url = $(dom_node).attr('data-results-url');
	
	this.expanded_pkgs = Array();
	
	
	this.update_packages = function () {
		/* Perform bits of setup with a new list of sourcepackages */
		var sourcepkgs = this.results_pane.find('li.sourcepackage');
		var old_expanded_pkgs = this.expanded_pkgs;
		
		this.expanded_pkgs = Array();
		
		sourcepkgs.each(function () {
			var pkg = $(this);
			
			/* handler for toggling a package's details on click */
			pkg.children('a.sourcepackage-header').bind('click', function () {
				/* Can we use $.load and abort when the user clicks again? */
				
				if ( pkg.hasClass('expanded') || pkg.hasClass('status-loading') ) {
					results.collapse_package(pkg);
				} else {
					results.expand_package(pkg);
				}
				
				return false;
			});
			
			pkg_id = $(this).attr('data-results-packageid');
			if ( pkg.hasClass('expanded') ) {
				results.expanded_pkgs.push(pkg_id);
			} else if ( old_expanded_pkgs.indexOf(pkg_id) > -1 ) {
				/* this package was expanded last time but we have
				   received it in collapsed form. It's likely the
				   results were requested before the user expanded it.
				   So, let's expand it again! */
				results.expand_package(pkg);
			}
		});
		
		this.update_current_query('expand_pkg', this.expanded_pkgs.join(','));
	}
	
	this.expand_package = function (pkg) {
		var pkg_id = pkg.attr('data-results-packageid');
		var pkg_details = pkg.children('.sourcepackage-details');
		
		var show = function () {
			pkg_details.animate({height: 'show', opacity: 'show'}, 150);
			pkg.removeClass('collapsed');
			pkg.addClass('expanded');
		}
		
		/* we record that the package is expanded first, so if anything
		   happens in between we can continue */
		this.expanded_pkgs.push(pkg_id);
		this.update_current_query('expand_pkg', this.expanded_pkgs.join(','));
		
		/* first, get package details from the server if they aren't already here */
		/* this poorly assumes that pkg_details contains no text at all if we haven't hit the server */
		if ( $.trim(pkg_details.html()) == '' ) {
			var pkg_url = this.results_url + '/' + pkg_id;
			pkg.addClass('status-loading');
			xhr = $.get(pkg_url, this.current_query, function (data, status, xhr) {
				/* finished loading; show the results */
				pkg_details.html(data);
				pkg.removeClass('status-loading');
				pkg.data('xhr', null);
				show();
			});
			pkg.data('xhr', xhr);
		} else {
			show();
		}
	}
	
	this.collapse_package = function (pkg) {
		var pkg_id = pkg.attr('data-results-packageid');
		var pkg_details = pkg.children('.sourcepackage-details');
		
		current_xhr = pkg.data('xhr');
		if ( current_xhr != null ) current_xhr.abort();
		
		array_index = this.expanded_pkgs.indexOf(pkg_id);
		this.expanded_pkgs.remove(array_index);
		this.update_current_query('expand_pkg', this.expanded_pkgs.join(','));
		
		pkg_details.animate({height: 'hide', opacity: 'hide'}, 150);
		/* Should we remove the now hidden inner html to save memory? */
		pkg.removeClass('expanded');
		pkg.addClass('collapsed');
	}
	
	
	this.load_results = function () {
		/* Begins loading results for future_query */
		
		var load_query = $.extend({}, this.current_query, this.future_query);
		
		current_xhr = this.results_pane.data('xhr');
		if ( current_xhr != null ) current_xhr.abort();
		
		this.results_pane.addClass('status-loading');
		
		xhr = $.get(this.results_url, load_query,
			function (data, status, xhr) {
				/* TODO: detect errors here and handle appropriately */
				
				/* display the (plain html) results we have receieved */
				results.results_pane.html(data);
				results.results_pane.removeClass('status-loading');
				
				results.current_query = load_query;
				results.future_query = {};
				
				results.results_pane.data('xhr', null);
				
				results.update_packages();
				
				/* display statistics in the footer */
				time_header = xhr.getResponseHeader('X-Django-Request-Time')
				if (time_header) {
					$('#requeststats').html('Results generated in '+parseFloat(time_header).toFixed(2) + ' seconds');
				}
			}
		);
		this.results_pane.data('xhr', xhr);
	}
	
	
	this.reset_timer = function () {
		/* Sets the query countdown timer for 1.5s.
		   When the timer ends, we load the results for future_query. */
		
		this.stop_timer();
		
		this.results_pane.addClass('status-waiting');
		
		this.query_countdown = setTimeout(
			function (results) {
				results.results_pane.removeClass('status-waiting');
				results.load_results();
			}, 1500, this);
	}
	
	this.stop_timer = function () {
		/* Stops the query countdown timer. */
		
		clearTimeout(this.query_countdown);
		this.query_countdown = 0;
		this.results_pane.removeClass('status-waiting');
	}
	
	
	this.post_query = function (key, value) {
		/* Adds the given key/value pair to future_query. If it is
		   necessary, a timer is started to get the results for
		   future_query. */
		
		if ( this.current_query[key] == value ) {
			/* the user has reverted a change; value is the same as before */
			delete this.future_query[key];
			
			if ( $.isEmptyObject(this.future_query) ) {
				/* no point getting results, so stop and wait for new input. */
				this.stop_timer();
			}
		} else {
			this.future_query[key] = value;
			/* start the timer from 0, waiting to apply the new future_query */
			this.reset_timer();
		}
	}
	
	this.update_current_query = function (key, value) {
		/* Adds the given key/value pair as part of current_query. For
		   example, when we first load the page it is useful to make
		   sure we know the current state of all the filters. */
		
		/* TODO: do something if there is an ongoing XHR request to make
		   sure we don't lose the new current_value... */
		this.current_query[key] = value;
	}
	
	
	/********/
	/* Object initialization */
	
	this.update_packages();
}



function create_filters () {
	/* Creates all filters for the document at once. */
	
	$('#filters').scrollFollow({ offset: 12 });
	
	var filters_list = new Array();
	
	$('#filters .filter').each(function (i) {
		filters_list.push(new Filter(this));
	});
	
	return filters_list;
}

function create_results () {
	return new Results($('#results'));
}



$(document).ready(function () {
	harvest_results = create_results();
	harvest_filters_list = create_filters();
	
	$('input[placeholder]').placeHeld();
});
